package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.time.LocalDate;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;

import org.apache.commons.lang.WordUtils;
import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.NotBlank;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.security.core.GrantedAuthority;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.VssUserDetails;
import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.shared.model.lookup.sds.Gender;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.OccasionalWorkEntry.OccasionalWorkEntryView;
import gov.va.vss.model.WorkEntry.WorkEntryView;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.lookup.Language;
import gov.va.vss.model.lookup.TransportationMethod;
import gov.va.vss.model.lookup.VolunteerStatus;
import gov.va.vss.model.organization.AbstractBasicOrganization;
import gov.va.vss.model.validation.constraints.ExtendedEmailValidator;
import gov.va.vss.util.DateUtil;
import gov.va.vss.util.ValidationUtil;

@Entity
@Table(name = "Volunteers", schema = "vss")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Volunteer extends AbstractAuditedVersionedPersistent<Volunteer>
		implements Comparable<Volunteer>, ObjectScopedToStationNumbers, VssUserDetails {
	private static final long serialVersionUID = 6904844123870655771L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class VolunteerView {
		public interface Basic {
		}

		public interface Search extends Basic {
		}

		public interface Extended extends Search {
		}
	}

	public static String getDisplayName(String firstName, String middleName, String lastName, String suffix, boolean lastFirst) {
		return WordUtils.capitalizeFully(StringUtil.getDisplayName(lastFirst, firstName, middleName, lastName, suffix));
	}
	
	public static String getDisplayName(String firstName, String middleName, String lastName, String suffix) {
		return getDisplayName(firstName, middleName, lastName, suffix, true);
	}

	// -------------------------------------- Fields

	@NotBlank
	private String lastName;
	@NotBlank
	private String firstName;
	private String middleName;
	private String suffix;
	@NotNull(message = "Please enter a gender.")
	private Gender gender;
	@NotNull(message = "Please enter a date of birth.")
	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate dateOfBirth;
	private String nickname;
	private Language preferredLanguage;
	private String remarks;
	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate entryDate;

	private String identifyingCode;
	private boolean vaEmployee;
	private String vaEmployeeUsername;

	private Integer lastAwardHours;
	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate lastAwardDate;
	private Award lastAward;

	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate statusDate;

	private VolunteerStatus status;

	private String terminationRemarks;
	private TransportationMethod transportationMethod;

	private Integer mealsEligible;
	private String mealRemarks;

	@NotBlank(message = "Please enter a street address.")
	private String addressLine1;
	private String addressLine2;
	@NotBlank(message = "Please enter a city.")
	private String city;
	@NotNull(message = "Please select a state.")
	private State state;
	@NotBlank(message = "Please enter a zipcode.")
	private String zip;
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String phone;
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String phoneAlt;
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String phoneAlt2;
	@ExtendedEmailValidator(message = "Please enter a valid email in the format 'user@domain.tld'.")
	private String email;
	private String emergencyContactName;
	private String emergencyContactRelationship;
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String emergencyContactPhone;
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String emergencyContactPhoneAlt;

	private List<VolunteerAssignment> volunteerAssignments;

	private Set<VolunteerOrganization> volunteerOrganizations;
	private AbstractBasicOrganization primaryOrganization;
	private Facility originallyCreatedAt;
	private Facility primaryFacility;
	private List<ParkingSticker> parkingStickers;
	private List<Uniform> uniforms;

	private String pivBadgeID;
	private LocalDate pivExpiration;

	private List<AdjustedHoursEntry> timeAdjustments;

	/* Mapping as list to ensure lazy loading - CPB */
	private List<Donor> donor;

	/* Computed */
	private Integer age;
	private boolean youth;
	private int grandfatheredYearsWorked;

	/* Transient security fields */
	private Collection<? extends GrantedAuthority> authorities = null;

	// -------------------------------------- Business Methods

	public Set<VolunteerAssignment> getAssignmentsByStatus(boolean activeStatus) {
		return getVolunteerAssignments().stream().filter(p -> activeStatus ? p.isActive() : p.isInactive())
				.collect(Collectors.toSet());
	}

	@Transient
	@Override
	@JsonIgnore
	public boolean isSoundsEnabled() {
		return false;
	}

	@Transient
	@JsonView(VolunteerView.Search.class)
	public String getAddressMultilineDisplay() {
		return getAddressDisplay(false);
	}

	private String getAddressDisplay(boolean useIdForState) {
		return StringUtil.getAddressDisplay(addressLine1, addressLine2, null, city,
				useIdForState ? String.valueOf(nullSafeGetId(getState())) : getState().getPostalName(), zip, "\n");
	}

	@Transient
	@JsonIgnore
	public Donor getDonor() {
		return getDonorList().isEmpty() ? null : getDonorList().get(0);
	}

	@Transient
	@JsonIgnore
	public Collection<String> getScopedToStationNumbers() {
		List<String> results = new ArrayList<>();
		for (VolunteerAssignment assignment : getVolunteerAssignments())
			results.add(assignment.getFacility().getStationNumber());
		return results;
	}

	@Transient
	public String getDisplayName(boolean lastFirst) {
		return getDisplayName(firstName, middleName, lastName, suffix, lastFirst);
	}
	
	@Transient
	@JsonView({ VolunteerView.Basic.class, WorkEntryView.TimeReportByDate.class,
			WorkEntryView.TimeReportByVolunteer.class, OccasionalWorkEntryView.TimeReport.class })
	public String getDisplayName() {
		return getDisplayName(true);
	}

	@Transient
	@JsonIgnore
	public Set<AbstractBasicOrganization> getBasicOrganizations() {
		Set<AbstractBasicOrganization> results = new HashSet<>();
		for (VolunteerOrganization o : getVolunteerOrganizations())
			results.add(o.getOrganization());
		return results;
	}

	public Set<VolunteerOrganization> getVolunteerOrganizationsByStatus(boolean activeStatus) {
		return getVolunteerOrganizations().stream().filter(p -> activeStatus ? p.isActive() : p.isInactive())
				.collect(Collectors.toSet());
	}

	@Override
	@Transient
	@JsonIgnore
	public Collection<? extends GrantedAuthority> getAuthorities() {
		return authorities;
	}

	@Override
	public Collection<? extends GrantedAuthority> getAuthoritiesAtFacility(long facilityId) {
		return authorities;
	}

	public void setAuthorities(Collection<? extends GrantedAuthority> authorities) {
		if (this.authorities != null)
			throw new IllegalStateException(
					"Cannot assign new authorities to volunteer once they have been assigned once.");
		this.authorities = authorities;
	}

	@Transient
	@Override
	@JsonIgnore
	public String getPassword() {
		LocalDate dob = getDateOfBirth();
		if (dob == null)
			return "_INVALID_password*#@$&(#@*)$&!!--1NKWE";
		return dob.format(DateUtil.TWO_DIGIT_DATE_ONLY_FORMAT);
	}

	@Transient
	@Override
	@JsonIgnore
	public String getUsername() {
		return getIdentifyingCode();
	}

	@Transient
	@Override
	@JsonIgnore
	public boolean isAccountNonExpired() {
		return true;
	}

	@Transient
	@Override
	@JsonIgnore
	public boolean isAccountNonLocked() {
		return true;
	}

	@Transient
	@Override
	@JsonIgnore
	public boolean isCredentialsNonExpired() {
		return true;
	}

	@Transient
	@Override
	@JsonIgnore
	public boolean isEnabled() {
		return !getStatus().isVolunteerTerminated();
	}

	@Transient
	@Override
	@JsonIgnore
	public ZoneId getTimeZone() {
		return null;
	}

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(Volunteer oo) {
		return new EqualsBuilder().append(getDisplayName(), oo.getDisplayName())
				.append(getDateOfBirth(), oo.getDateOfBirth()).append(getIdentifyingCode(), oo.getIdentifyingCode())
				.isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getDisplayName()).append(getDateOfBirth()).append(getIdentifyingCode())
				.toHashCode();
	}

	@Override
	public int compareTo(Volunteer o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(nullSafeLowercase(getDisplayName()), nullSafeLowercase(o.getDisplayName()))
				.toComparison() > 0 ? 1 : -1;
	}

	public String toString() {
		return getDisplayName() + " (DOB " + getDateOfBirth() + ", code '" + identifyingCode + "')";
	}

	// -------------------------------------- Accessor Methods

	@Column(length = 30)
	@JsonView(VolunteerView.Extended.class)
	public String getLastName() {
		return lastName;
	}

	public void setLastName(String lastName) {
		this.lastName = lastName;
	}

	@Column(length = 30)
	@JsonView(VolunteerView.Extended.class)
	public String getFirstName() {
		return firstName;
	}

	public void setFirstName(String firstName) {
		this.firstName = firstName;
	}

	@Column(length = 20)
	@JsonView(VolunteerView.Extended.class)
	public String getMiddleName() {
		return middleName;
	}

	public void setMiddleName(String middleName) {
		this.middleName = middleName;
	}

	@Column(name = "NameSuffix", length = 10)
	@JsonView(VolunteerView.Extended.class)
	public String getSuffix() {
		return suffix;
	}

	public void setSuffix(String suffix) {
		this.suffix = suffix;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "STD_GenderFK")
	@JsonView(VolunteerView.Search.class)
	public Gender getGender() {
		return gender;
	}

	public void setGender(Gender gender) {
		this.gender = gender;
	}

	@JsonView(VolunteerView.Basic.class)
	public LocalDate getDateOfBirth() {
		return dateOfBirth;
	}

	public void setDateOfBirth(LocalDate dateOfBirth) {
		this.dateOfBirth = dateOfBirth;
	}

	@Column(name = "NickName", length = 30)
	@JsonView(VolunteerView.Extended.class)
	public String getNickname() {
		return nickname;
	}

	public void setNickname(String nickname) {
		this.nickname = nickname;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "PreferredLanguageForLoginFK")
	@JsonView(VolunteerView.Extended.class)
	public Language getPreferredLanguage() {
		return preferredLanguage;
	}

	public void setPreferredLanguage(Language preferredLanguage) {
		this.preferredLanguage = preferredLanguage;
	}

	@JsonView(VolunteerView.Extended.class)
	public String getRemarks() {
		return remarks;
	}

	public void setRemarks(String remarks) {
		this.remarks = remarks;
	}

	@Column(name = "TerminatedRemarks")
	@JsonView(VolunteerView.Extended.class)
	public String getTerminationRemarks() {
		return terminationRemarks;
	}

	public void setTerminationRemarks(String terminationRemarks) {
		this.terminationRemarks = terminationRemarks;
	}

	@Column(name = "VAEmployee", nullable = false)
	@JsonView(VolunteerView.Extended.class)
	public boolean isVaEmployee() {
		return vaEmployee;
	}

	public void setVaEmployee(boolean vaEmployee) {
		this.vaEmployee = vaEmployee;
	}

	@JsonView(VolunteerView.Extended.class)
	public String getVaEmployeeUsername() {
		return vaEmployeeUsername;
	}

	public void setVaEmployeeUsername(String vaEmployeeUsername) {
		this.vaEmployeeUsername = vaEmployeeUsername;
	}

	@Column(insertable = false, updatable = false)
	@JsonView(VolunteerView.Extended.class)
	public Integer getAge() {
		return age;
	}

	public void setAge(Integer age) {
		this.age = age;
	}
	
	@Column(name = "IsYouth", insertable = false, updatable = false, nullable = false)
	public boolean isYouth() {
		return youth;
	}

	public void setYouth(boolean youth) {
		this.youth = youth;
	}

	@Column(insertable = false, updatable = false)
	@JsonView(VolunteerView.Extended.class)
	public int getGrandfatheredYearsWorked() {
		return grandfatheredYearsWorked;
	}

	public void setGrandfatheredYearsWorked(int grandfatheredYearsWorked) {
		this.grandfatheredYearsWorked = grandfatheredYearsWorked;
	}

	@Column(length = 10)
	@JsonView(VolunteerView.Search.class)
	public String getIdentifyingCode() {
		return identifyingCode;
	}

	public void setIdentifyingCode(String identifyingCode) {
		this.identifyingCode = identifyingCode;
	}

	@Column(name = "StreetAddress1", length = 35, nullable = false)
	@JsonView(VolunteerView.Extended.class)
	public String getAddressLine1() {
		return addressLine1;
	}

	public void setAddressLine1(String addressLine1) {
		this.addressLine1 = addressLine1;
	}

	@Column(name = "StreetAddress2", length = 35)
	@JsonView(VolunteerView.Extended.class)
	public String getAddressLine2() {
		return addressLine2;
	}

	public void setAddressLine2(String addressLine2) {
		this.addressLine2 = addressLine2;
	}

	@Column(length = 30, nullable = false)
	@JsonView(VolunteerView.Extended.class)
	public String getCity() {
		return city;
	}

	public void setCity(String city) {
		this.city = city;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "STD_StateFK", nullable = false)
	@JsonView(VolunteerView.Extended.class)
	public State getState() {
		return state;
	}

	public void setState(State state) {
		this.state = state;
	}

	@Column(name = "ZipCode", length = 10)
	@JsonView(VolunteerView.Extended.class)
	public String getZip() {
		return zip;
	}

	public void setZip(String zip) {
		this.zip = zip;
	}

	@Column(name = "Telephone", length = 30)
	@JsonView(VolunteerView.Search.class)
	public String getPhone() {
		return phone;
	}

	public void setPhone(String phone) {
		this.phone = phone;
	}

	@Column(name = "AlternateTelephone", length = 30)
	@JsonView(VolunteerView.Extended.class)
	public String getPhoneAlt() {
		return phoneAlt;
	}

	public void setPhoneAlt(String phoneAlternate) {
		this.phoneAlt = phoneAlternate;
	}

	@Column(name = "AlternateTelephone2", length = 30)
	@JsonView(VolunteerView.Extended.class)
	public String getPhoneAlt2() {
		return phoneAlt2;
	}

	public void setPhoneAlt2(String phoneAlternate2) {
		this.phoneAlt2 = phoneAlternate2;
	}

	@Column(name = "EMailAddress", length = 250)
	@JsonView(VolunteerView.Search.class)
	public String getEmail() {
		return email;
	}

	public void setEmail(String email) {
		this.email = email;
	}

	@Column(name = "HoursLastAward")
	@JsonView(VolunteerView.Extended.class)
	public Integer getLastAwardHours() {
		return lastAwardHours;
	}

	public void setLastAwardHours(Integer lastAwardHours) {
		this.lastAwardHours = lastAwardHours;
	}

	@Column(name = "DateLastAward")
	@JsonView(VolunteerView.Extended.class)
	public LocalDate getLastAwardDate() {
		return lastAwardDate;
	}

	public void setLastAwardDate(LocalDate lastAwardDate) {
		this.lastAwardDate = lastAwardDate;
	}

	@ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.ALL)
	@JoinColumn(name = "AwardCodesFK")
	@JsonView(VolunteerView.Extended.class)
	public Award getLastAward() {
		return lastAward;
	}

	public void setLastAward(Award lastAward) {
		this.lastAward = lastAward;
	}

	@JsonView(VolunteerView.Search.class)
	@Column(insertable = false, updatable = false)
	public LocalDate getStatusDate() {
		return statusDate;
	}

	public void setStatusDate(LocalDate statusDate) {
		this.statusDate = statusDate;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VSS_STD_TransportationMethodFK")
	@JsonView(VolunteerView.Extended.class)
	public TransportationMethod getTransportationMethod() {
		return transportationMethod;
	}

	public void setTransportationMethod(TransportationMethod transportationMethod) {
		this.transportationMethod = transportationMethod;
	}

	@Column(name = "EligibleForNumMeals")
	@JsonView(VolunteerView.Extended.class)
	public Integer getMealsEligible() {
		return mealsEligible;
	}

	public void setMealsEligible(Integer mealsEligible) {
		this.mealsEligible = mealsEligible;
	}

	@Column(length = 250, name = "EmergencyContactName")
	@JsonView(VolunteerView.Extended.class)
	public String getEmergencyContactName() {
		return emergencyContactName;
	}

	public void setEmergencyContactName(String emergencyContactName) {
		this.emergencyContactName = emergencyContactName;
	}

	@Column(length = 250, name = "EmergencyContactRelationship")
	@JsonView(VolunteerView.Extended.class)
	public String getEmergencyContactRelationship() {
		return emergencyContactRelationship;
	}

	public void setEmergencyContactRelationship(String emergencyContactRelationship) {
		this.emergencyContactRelationship = emergencyContactRelationship;
	}

	@Column(length = 30, name = "EmergencyContactTelephone")
	@JsonView(VolunteerView.Extended.class)
	public String getEmergencyContactPhone() {
		return emergencyContactPhone;
	}

	public void setEmergencyContactPhone(String emergencyContactPhone) {
		this.emergencyContactPhone = emergencyContactPhone;
	}

	@Column(length = 30, name = "EmergencyContactAlternateTelephone")
	@JsonView(VolunteerView.Extended.class)
	public String getEmergencyContactPhoneAlt() {
		return emergencyContactPhoneAlt;
	}

	public void setEmergencyContactPhoneAlt(String emergencyContactPhoneAlt) {
		this.emergencyContactPhoneAlt = emergencyContactPhoneAlt;
	}

	@OneToMany(mappedBy = "volunteer", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public Set<VolunteerOrganization> getVolunteerOrganizations() {
		if (volunteerOrganizations == null)
			volunteerOrganizations = new HashSet<>();
		return volunteerOrganizations;
	}

	public void setVolunteerOrganizations(Set<VolunteerOrganization> volunteerOrganizations) {
		this.volunteerOrganizations = volunteerOrganizations;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "PrimaryOrganizationFK", insertable = false, updatable = false)
	@BatchSize(size = 500)
	@JsonIgnore
	public AbstractBasicOrganization getPrimaryOrganization() {
		return primaryOrganization;
	}

	/* Managed by a native method in VolunteerDAOImpl */
	@SuppressWarnings("unused")
	private void setPrimaryOrganization(AbstractBasicOrganization primaryOrganization) {
		this.primaryOrganization = primaryOrganization;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "PrimaryFacilityFK", insertable = false, updatable = false)
	@BatchSize(size = 500)
	@JsonView(VolunteerView.Basic.class)
	public Facility getPrimaryFacility() {
		return primaryFacility;
	}

	/* Managed by a native method in VolunteerDAOImpl */
	@SuppressWarnings("unused")
	private void setPrimaryFacility(Facility primaryFacility) {
		this.primaryFacility = primaryFacility;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "OriginalFacilityCreatedFK", nullable = false)
	@BatchSize(size = 500)
	@JsonView(VolunteerView.Extended.class)
	public Facility getOriginallyCreatedAt() {
		return originallyCreatedAt;
	}

	public void setOriginallyCreatedAt(Facility originallyCreatedAt) {
		this.originallyCreatedAt = originallyCreatedAt;
	}

	@OneToMany(mappedBy = "volunteer", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<VolunteerAssignment> getVolunteerAssignments() {
		if (volunteerAssignments == null)
			volunteerAssignments = new ArrayList<>();
		return volunteerAssignments;
	}

	public void setVolunteerAssignments(List<VolunteerAssignment> volunteerAssignments) {
		this.volunteerAssignments = volunteerAssignments;
	}

	@JsonView(VolunteerView.Extended.class)
	public LocalDate getEntryDate() {
		return entryDate;
	}

	public void setEntryDate(LocalDate entryDate) {
		this.entryDate = entryDate;
	}

	@OneToMany(mappedBy = "volunteer", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<ParkingSticker> getParkingStickers() {
		if (parkingStickers == null)
			parkingStickers = new ArrayList<>();
		return parkingStickers;
	}

	public void setParkingStickers(List<ParkingSticker> parkingStickers) {
		this.parkingStickers = parkingStickers;
	}

	@OneToMany(mappedBy = "volunteer", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<Uniform> getUniforms() {
		if (uniforms == null)
			uniforms = new ArrayList<>();
		return uniforms;
	}

	public void setUniforms(List<Uniform> uniforms) {
		this.uniforms = uniforms;
	}

	@Column(name = "PIVBadgeID", length = 100)
	@JsonView(VolunteerView.Extended.class)
	public String getPivBadgeID() {
		return pivBadgeID;
	}

	public void setPivBadgeID(String pivBadgeID) {
		this.pivBadgeID = pivBadgeID;
	}

	@Column(name = "PIVExpiration")
	@JsonView(VolunteerView.Extended.class)
	public LocalDate getPivExpiration() {
		return pivExpiration;
	}

	public void setPivExpiration(LocalDate pivExpiration) {
		this.pivExpiration = pivExpiration;
	}

	@JsonIgnore
	public String getMealRemarks() {
		return mealRemarks;
	}

	public void setMealRemarks(String mealRemarks) {
		this.mealRemarks = mealRemarks;
	}

	@OneToMany(mappedBy = "volunteer", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<AdjustedHoursEntry> getTimeAdjustments() {
		if (timeAdjustments == null)
			timeAdjustments = new ArrayList<>();
		return timeAdjustments;
	}

	public void setTimeAdjustments(List<AdjustedHoursEntry> timeAdjustments) {
		this.timeAdjustments = timeAdjustments;
	}

	@OneToMany(mappedBy = "volunteer", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	private List<Donor> getDonorList() {
		if (donor == null)
			donor = new ArrayList<>();
		return donor;
	}

	public void setDonorList(List<Donor> donor) {
		this.donor = donor;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VSS_STD_VolunteerStatusFK", nullable = false, insertable = false, updatable = false)
	@BatchSize(size = 500)
	@JsonView(VolunteerView.Search.class)
	public VolunteerStatus getStatus() {
		return status;
	}

	public void setStatus(VolunteerStatus status) {
		this.status = status;
	}

}
