package gov.va.vss.model.requirement;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.DiscriminatorColumn;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.Inheritance;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.Table;
import javax.persistence.Transient;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.Hibernate;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.NotBlank;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.lookup.RequirementDateType;
import gov.va.vss.web.conversion.RequirementDateTypeConverter;

@Entity
@Inheritance
@DiscriminatorColumn(name = "Type")
@Table(name = "Requirement", schema = "vss")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public abstract class AbstractRequirement extends AbstractAuditedVersionedPersistent<AbstractRequirement>
		implements Comparable<AbstractRequirement> {
	private static final long serialVersionUID = 583796042812902141L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class RequirementView {
		public interface Basic {
		}

		public interface Extended extends Basic {
		}
	}

	// -------------------------------------- Fields

	@Length(max = 50)
	@NotBlank(message = "Name is required.")
	private String name;
	@Length(max = 250)
	private String description;
	private boolean inactive;
	private RequirementDateType dateType;

	// -------------------------------------- Business Methods

	@Transient
	public abstract Facility getFacilityScope();

	// -------------------------------------- Common Methods

	@Transient
	private int getClassOrder() {
		Class<?> clazz = Hibernate.getClass(this);
		if (clazz == GlobalRequirement.class)
			return 0;
		if (clazz == GlobalRoleRequirement.class)
			return 1;
		if (clazz == FacilityRequirement.class)
			return 2;
		if (clazz == FacilityRoleRequirement.class)
			return 3;
		throw new AssertionError("Unexpected instance of this class: " + clazz);
	}

	@Override
	public int compareTo(AbstractRequirement o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(getClassOrder(), o.getClassOrder())
				.append(nullSafeLowercase(getName()), nullSafeLowercase(o.getName())).toComparison() > 0 ? 1 : -1;
	}

	@Override
	protected boolean requiredEquals(AbstractRequirement oo) {
		return new EqualsBuilder().append(getName(), oo.getName())
				.append(nullSafeGetId(getFacilityScope()), nullSafeGetId(oo.getFacilityScope())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getName()).append(nullSafeGetId(getFacilityScope())).toHashCode();
	}

	public String toString() {
		return getName();
	}

	// -------------------------------------- Accessor Methods

	@Column(length = 50, nullable = false)
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	@Column(length = 250)
	@JsonView(RequirementView.Basic.class)
	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	@Column(name = "IsInactive", nullable = false)
	@JsonView(RequirementView.Basic.class)
	public boolean isInactive() {
		return inactive;
	}

	public void setInactive(boolean inactive) {
		this.inactive = inactive;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VSS_STD_RequirementDateTypeFK", unique = true)
	@Convert(converter = RequirementDateTypeConverter.class)
	@JsonView(RequirementView.Basic.class)
	public RequirementDateType getDateType() {
		return dateType;
	}

	public void setDateType(RequirementDateType dateType) {
		this.dateType = dateType;
	}

	
}