package gov.va.vss.service.impl;

import static gov.va.shared.util.StringUtil.normalizeLineBreaks;
import static gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType.ACTIVE;
import static gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType.INACTIVE;
import static org.apache.commons.lang.WordUtils.capitalizeFully;
import static org.apache.commons.lang3.StringUtils.trim;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.commons.lang.WordUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Service;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.Permission;
import gov.va.shared.service.validation.ServiceValidationException;
import gov.va.vss.model.AvailableIdentifyingCode;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.VolunteerAssignment;
import gov.va.vss.model.VolunteerOrganization;
import gov.va.vss.model.WorkEntry;
import gov.va.vss.model.benefitingService.BenefitingService;
import gov.va.vss.model.benefitingService.BenefitingServiceRole;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.lookup.Language;
import gov.va.vss.model.lookup.VolunteerStatus;
import gov.va.vss.model.lookup.Language.LanguageType;
import gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType;
import gov.va.vss.model.organization.AbstractBasicOrganization;
import gov.va.vss.model.organization.ScopeType;
import gov.va.vss.service.VolunteerService;
import gov.va.vss.service.requirement.VolunteerRequirementService;
import gov.va.vss.util.DateUtil;

@Service
public class VolunteerServiceImpl extends AbstractServiceImpl implements VolunteerService {
	private static final Logger log = LoggerFactory.getLogger(VolunteerServiceImpl.class);

	@Value("${maxIdleDaysBeforeVolunteerInactivation}")
	private int maxIdleDaysBeforeVolunteerInactivation;
	@Value("${volunteerInactivationGracePeriod}")
	private int volunteerInactivationGracePeriod;

	@Autowired
	private VolunteerRequirementService volunteerRequirementService;

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public Volunteer saveOrUpdate(Volunteer volunteer, AppUser userContext) throws ServiceValidationException {
		boolean isNew = !volunteer.isPersistent();

		Volunteer existingVolunteer = isNew ? null : volunteerDAO.findRequiredByPrimaryKey(volunteer.getId());

		boolean appendUserToMainRemarks = StringUtils.isNotBlank(volunteer.getRemarks()) && (isNew
				|| existingVolunteer.getRemarks() == null || !normalizeLineBreaks(existingVolunteer.getRemarks())
						.equals(normalizeLineBreaks(volunteer.getRemarks())));
		boolean appendUserToMealsRemarks = StringUtils.isNotBlank(volunteer.getMealRemarks())
				&& (isNew || existingVolunteer.getMealRemarks() == null
						|| !normalizeLineBreaks(existingVolunteer.getMealRemarks())
								.equals(normalizeLineBreaks(volunteer.getMealRemarks())));
		boolean appendUserToTerminationRemarks = StringUtils.isNotBlank(volunteer.getTerminationRemarks())
				&& (isNew || existingVolunteer.getTerminationRemarks() == null
						|| !normalizeLineBreaks(existingVolunteer.getTerminationRemarks())
								.equals(normalizeLineBreaks(volunteer.getTerminationRemarks())));

		if (isNew) {
			volunteer.setEntryDate(getTodayAtFacility());
			volunteer.setOriginallyCreatedAt(getFacilityContext());
			volunteer.setPreferredLanguage(languageDAO.findByLookup(LanguageType.ENGLISH));
			AvailableIdentifyingCode c = availableIdentifyingCodeDAO.getFirstUnused();
			volunteer.setIdentifyingCode(c.getCode());
			availableIdentifyingCodeDAO.delete(c);
		}
		volunteer.setFirstName(trim(capitalizeFully(volunteer.getFirstName())));
		volunteer.setMiddleName(trim(capitalizeFully(volunteer.getMiddleName())));
		volunteer.setLastName(trim(capitalizeFully(volunteer.getLastName())));
		volunteer.setSuffix(trim(capitalizeFully(volunteer.getSuffix())));
		volunteer.setZip(trim(volunteer.getZip()));

		if (appendUserToMainRemarks)
			volunteer.setRemarks(volunteer.getRemarks() + getAuditText(userContext));
		if (appendUserToMealsRemarks)
			volunteer.setMealRemarks(volunteer.getMealRemarks() + getAuditText(userContext));
		if (appendUserToTerminationRemarks)
			volunteer.setTerminationRemarks(volunteer.getTerminationRemarks() + getAuditText(userContext));

		Volunteer v = volunteerDAO.saveOrUpdate(volunteer);

		if (isNew) {
			/* Set the primary facility and initial status */
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, true, getFacilityContext().getId(),
					VolunteerStatusType.ACTIVE, getTodayAtFacility(), null);
			volunteerRequirementDAO.bulkInsertForVolunteerIfNecessary(v.getId(), getFacilityContext().getId());
		} else {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, false, null,
					v.getStatus().getLookupType(), v.getStatusDate(), null);

			if (v.getStatus().isVolunteerInactiveOrTerminated()) {
				for (VolunteerAssignment va : v.getVolunteerAssignments()) {
					if (va.isActive()) {
						va.setInactive(true);
						va = volunteerAssignmentDAO.saveOrUpdate(va);
					}
				}
				for (VolunteerOrganization vo : v.getVolunteerOrganizations()) {
					if (vo.isActive()) {
						vo.setInactive(true);
						vo = volunteerOrganizationDAO.saveOrUpdate(vo);
					}
				}
			}

			if (!v.getStatus().isVolunteerTerminated())
				v = updateVolunteerStatusIfNeeded(v).updatedVolunteer;
		}

		return v;
	}

	private String getAuditText(AppUser appUser) {
		ZoneId timeZone = appUser.getTimeZone();
		if (timeZone == null)
			timeZone = ZoneId.systemDefault();

		return "\n - " + appUser.getDisplayName() + " [" + appUser.getUsername() + "] @"
				+ ZonedDateTime.now(timeZone).format(DateUtil.MILITARY_DATE_TIME_WITH_ZONE_FORMAT);
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public void delete(long volunteerId) {
		volunteerDAO.delete(volunteerId);
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public void setPrimaryOrganization(long volunteerId, long organizationId) {
		volunteerDAO.updateFieldsWithoutVersionIncrement(volunteerId, true, organizationId, false, null, null, null,
				null);
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public void setPrimaryFacility(long volunteerId, long facilityId) {
		volunteerDAO.updateFieldsWithoutVersionIncrement(volunteerId, false, null, true, facilityId, null, null, null);
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public boolean addOrReactivateOrganization(long volunteerId, long organizationId)
			throws ServiceValidationException {
		List<VolunteerOrganization> list = volunteerOrganizationDAO.findByCriteria(volunteerId, organizationId, null,
				null);
		VolunteerOrganization vo = null;
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		if (v.getStatus().isVolunteerInactiveOrTerminated())
			throw new ServiceValidationException(
					"volunteer.error.organization.addOrganizationToInactiveOrTerminatedVolunteer");

		if (list.isEmpty()) {
			AbstractBasicOrganization o = organizationDAO.findRequiredByPrimaryKey(organizationId);
			if (o.isInactive())
				throw new ServiceValidationException(o.getScope() == ScopeType.NATIONAL
						? "volunteer.error.organization.addInactiveNationalOrganization"
						: "volunteer.error.organization.addInactiveLocalOrganization");
			vo = new VolunteerOrganization(v, o);
		} else {
			vo = list.get(0);
			AbstractBasicOrganization o = vo.getOrganization();
			if (o.isInactive())
				throw new ServiceValidationException(o.getScope() == ScopeType.NATIONAL
						? "volunteer.error.organization.reactivateInactiveNationalOrganization"
						: "volunteer.error.organization.reactivateInactiveLocalOrganization");
			vo.setInactive(false);
		}

		vo = volunteerOrganizationDAO.saveOrUpdate(vo);

		VolunteerStatusUpdateResult r = updateVolunteerStatusIfNeeded(v);

		if (v.getPrimaryOrganization() == null)
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), true, organizationId, false, null, null, null,
					null);

		return r.statusChanged;
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public Map<String, Object> inactivateOrganization(long volunteerOrganizationId) {
		VolunteerOrganization vo = volunteerOrganizationDAO.findRequiredByPrimaryKey(volunteerOrganizationId);
		Volunteer v = vo.getVolunteer();

		AbstractBasicOrganization currentPrimary = v.getPrimaryOrganization();
		AbstractBasicOrganization orgBeingInactivated = vo.getOrganization();

		boolean primaryWasInactivatedOrDeleted = orgBeingInactivated.equals(currentPrimary);

		vo.setInactive(true);
		vo = volunteerOrganizationDAO.saveOrUpdate(vo);

		return updateStatusAndPrimaryOrgIfNeeded(v, primaryWasInactivatedOrDeleted);
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public Map<String, Object> deleteOrganization(long volunteerOrganizationId) {
		VolunteerOrganization vo = volunteerOrganizationDAO.findRequiredByPrimaryKey(volunteerOrganizationId);
		Volunteer v = vo.getVolunteer();

		AbstractBasicOrganization currentPrimary = v.getPrimaryOrganization();
		AbstractBasicOrganization orgBeingInactivated = vo.getOrganization();

		boolean primaryWasInactivatedOrDeleted = orgBeingInactivated.equals(currentPrimary);

		List<WorkEntry> workEntries = workEntryDAO.findByCriteria(v.getId(), null, null, orgBeingInactivated.getId(),
				null, null);
		if (workEntries.isEmpty()) {
			volunteerOrganizationDAO.delete(volunteerOrganizationId);
		} else {
			vo.setInactive(true);
			vo = volunteerOrganizationDAO.saveOrUpdate(vo);
		}

		return updateStatusAndPrimaryOrgIfNeeded(v, primaryWasInactivatedOrDeleted);
	}

	private Map<String, Object> updateStatusAndPrimaryOrgIfNeeded(Volunteer v, boolean primaryWasInactivatedOrDeleted) {
		volunteerDAO.flushAndRefresh(v);

		VolunteerStatusUpdateResult r = updateVolunteerStatusIfNeeded(v);

		boolean reportPrimaryWasInactivated = false;

		Set<VolunteerOrganization> activeOrgAssignments = v.getVolunteerOrganizationsByStatus(true);
		if (activeOrgAssignments.size() == 1) {
			/*
			 * If we deleted or inactivated an org, and afterward there's only
			 * one active one left, this must become our new primary in all
			 * cases. Select it and don't bother notifying the user.
			 */
			AbstractBasicOrganization organization = activeOrgAssignments.iterator().next().getOrganization();
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), true, organization.getId(), false, null, null,
					null, null);
		} else if (primaryWasInactivatedOrDeleted) {
			/*
			 * If we inactivated/deleted our current primary, and afterward
			 * there's either 0 or more than 1 active orgs left, null out the
			 * primary value. Only report that a new one will need to be
			 * selected if one is available for selection.
			 */
			reportPrimaryWasInactivated = !activeOrgAssignments.isEmpty();
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), true, null, false, null, null, null, null);
		}

		Map<String, Object> results = new HashMap<>();
		results.put("primaryOrgInactivated", reportPrimaryWasInactivated);
		results.put("volunteerStatusChanged", r.statusChanged);
		return results;
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public boolean inactivateAssignment(long volunteerAssignmentId) {
		VolunteerAssignment vo = volunteerAssignmentDAO.findRequiredByPrimaryKey(volunteerAssignmentId);
		Volunteer v = vo.getVolunteer();
		Map<Long, LocalDate> workEntries = workEntryDAO.getMostRecentVolunteeredDateByFacilityMap(v.getId());

		Facility currentPrimary = v.getPrimaryFacility();

		vo.setInactive(true);
		vo = volunteerAssignmentDAO.saveOrUpdate(vo);

		boolean statusChanged = updatePrimaryFacilityAndStatus(v, workEntries, currentPrimary);
		return statusChanged;
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public boolean deleteAssignment(long volunteerAssignmentId) {
		VolunteerAssignment vo = volunteerAssignmentDAO.findRequiredByPrimaryKey(volunteerAssignmentId);
		Volunteer v = vo.getVolunteer();
		Map<Long, LocalDate> workEntries = workEntryDAO.getMostRecentVolunteeredDateByFacilityMap(v.getId());

		Facility currentPrimary = v.getPrimaryFacility();

		if (vo.getWorkEntries().isEmpty()) {
			volunteerAssignmentDAO.delete(volunteerAssignmentId);
		} else {
			vo.setInactive(true);
			vo = volunteerAssignmentDAO.saveOrUpdate(vo);
		}

		boolean statusChanged = updatePrimaryFacilityAndStatus(v, workEntries, currentPrimary);
		return statusChanged;
	}

	private boolean updatePrimaryFacilityAndStatus(Volunteer v, Map<Long, LocalDate> workEntries,
			Facility currentPrimary) {
		volunteerDAO.flushAndRefresh(v);

		SortedMap<LocalDate, Facility> facilityPrioritizedChoices = new TreeMap<>();

		/* Get the active facility with the most recent hours worked */
		Set<VolunteerAssignment> activeAssignments = v.getAssignmentsByStatus(true);
		for (VolunteerAssignment a : activeAssignments) {
			Facility f = a.getFacility().getFacility();
			/*
			 * If we have another active assignment at our current primary
			 * facility, no work needed - return - CPB
			 */
			if (f.equals(currentPrimary))
				return false;

			LocalDate mostRecentWorkDate = workEntries.get(f.getId());
			if (mostRecentWorkDate != null)
				facilityPrioritizedChoices.put(mostRecentWorkDate, f);
		}

		/*
		 * If we don't have any active assignments, get the inactive assignment
		 * facility with the most recent hours worked - CPB
		 */
		if (facilityPrioritizedChoices.isEmpty()) {
			Set<VolunteerAssignment> inactiveAssignments = v.getAssignmentsByStatus(false);
			for (VolunteerAssignment a : inactiveAssignments) {
				Facility f = a.getFacility().getFacility();
				LocalDate mostRecentWorkDate = workEntries.get(f.getId());
				if (mostRecentWorkDate != null)
					facilityPrioritizedChoices.put(mostRecentWorkDate, f);
			}
		}

		VolunteerStatusUpdateResult r = updateVolunteerStatusIfNeeded(v);

		if (facilityPrioritizedChoices.isEmpty()) {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, true,
					v.getOriginallyCreatedAt() == null ? null : v.getOriginallyCreatedAt().getId(), null, null, null);
		} else {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, true,
					facilityPrioritizedChoices.get(facilityPrioritizedChoices.lastKey()).getId(), null, null, null);
		}

		return r.statusChanged;
	}

	@Override
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public boolean addOrReactivateAssignment(Long volunteerAssignmentId, Long volunteerId, long facilityId,
			Long benefitingServiceRoleId) throws ServiceValidationException {
		VolunteerAssignment va = null;
		Volunteer v = null;

		if (volunteerAssignmentId != null) {
			va = volunteerAssignmentDAO.findRequiredByPrimaryKey(volunteerAssignmentId);
			v = va.getVolunteer();

			/* Needed for the bulk insert query below - CPB */
			volunteerId = v.getId();

			if (v.getStatus().isVolunteerTerminated())
				throw new ServiceValidationException("volunteerAssignment.reactivate.errorVolunteerTerminated");

			BenefitingService bs = va.getBenefitingService();
			if (bs.isInactive()) {
				if (bs.isNational() && bs.getTemplate().isInactive())
					throw new ServiceValidationException(
							"volunteerAssignment.reactivate.errorNationalBenefitingServiceInactive");

				throw new ServiceValidationException(
						"volunteerAssignment.reactivate.errorLocalBenefitingServiceInactive", bs.getDisplayName());
			}

			BenefitingServiceRole bsr = va.getBenefitingServiceRole();
			if (bsr.isInactive()) {
				if (bsr.isNational() && bsr.getTemplate().isInactive())
					throw new ServiceValidationException(
							"volunteerAssignment.reactivate.errorNationalBenefitingServiceInactive");

				throw new ServiceValidationException(
						"volunteerAssignment.reactivate.errorLocalBenefitingServiceInactive", bs.getDisplayName());
			}

			va.setInactive(false);
			va = volunteerAssignmentDAO.saveOrUpdate(va);
		} else {
			v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
			if (v.getStatus().isVolunteerTerminated())
				throw new ServiceValidationException("volunteerAssignment.reactivate.errorVolunteerTerminated");

			// prevent duplicate {volunteer, benefiting service role, location}
			List<VolunteerAssignment> potentialDuplicates = volunteerAssignmentDAO.findByCriteria(volunteerId, null,
					benefitingServiceRoleId, facilityId, null);
			if (!potentialDuplicates.isEmpty())
				throw new ServiceValidationException("volunteerAssignment.error.duplicateAssignmentFound");

			BenefitingServiceRole bsr = benefitingServiceRoleDAO.findRequiredByPrimaryKey(benefitingServiceRoleId);
			BenefitingService bs = bsr.getBenefitingService();

			va = new VolunteerAssignment();
			va.setBenefitingService(bs);
			va.setBenefitingServiceRole(bsr);
			va.setFacility(bsr.getFacility());
			va.setVolunteer(v);
			va = volunteerAssignmentDAO.saveOrUpdate(va);
		}

		VolunteerStatusUpdateResult r = updateVolunteerStatusIfNeeded(v);
		volunteerRequirementService.bulkInsertForVolunteerIfNecessary(volunteerId, facilityId);

		return r.statusChanged;
	}

	static class VolunteerStatusUpdateResult {
		Volunteer updatedVolunteer;
		boolean statusChanged;

		public VolunteerStatusUpdateResult(Volunteer updatedVolunteer, boolean statusChanged) {
			this.updatedVolunteer = updatedVolunteer;
			this.statusChanged = statusChanged;
		}
	}

	private VolunteerStatusUpdateResult updateVolunteerStatusIfNeeded(Volunteer v) {
		LocalDate cutoffDate = getTodayAtFacility().minusDays(maxIdleDaysBeforeVolunteerInactivation);

		boolean hasActiveAssignments = !v.getAssignmentsByStatus(true).isEmpty();
		boolean hasInactiveAssignments = !v.getAssignmentsByStatus(false).isEmpty();

		boolean changed = false;
		VolunteerStatus status = v.getStatus();

		VolunteerStatusType correctType = null;
		if (v.getEntryDate().isAfter(cutoffDate)) {
			if (hasActiveAssignments) {
				correctType = ACTIVE;
			} else if (hasInactiveAssignments) {
				correctType = INACTIVE;
			} else {
				correctType = ACTIVE;
			}
		} else {
			correctType = hasActiveAssignments ? ACTIVE : INACTIVE;
		}

		if (status.getLookupType() == INACTIVE && correctType == ACTIVE) {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, false, null, ACTIVE,
					getTodayAtFacility(), null);
			changed = true;
		} else if (status.getLookupType() == ACTIVE && correctType == INACTIVE) {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, false, null, INACTIVE,
					getTodayAtFacility(), null);

			for (VolunteerOrganization o : v.getVolunteerOrganizations()) {
				if (o.isActive()) {
					o.setInactive(true);
					o = volunteerOrganizationDAO.saveOrUpdate(o);
				}
			}

			changed = true;
		}
		return new VolunteerStatusUpdateResult(v, changed);
	}

	@Override
	public int inactivateStaleVolunteers(ZoneId switchToTimeZone) {
		ZonedDateTime cutoffDate = ZonedDateTime.now(switchToTimeZone)
				.minusDays(maxIdleDaysBeforeVolunteerInactivation);
		ZonedDateTime gracePeriodCutoff = ZonedDateTime.now(switchToTimeZone)
				.minusDays(volunteerInactivationGracePeriod);
		volunteerAssignmentDAO.inactivateStaleAssignments(cutoffDate, gracePeriodCutoff);
		int updated = volunteerDAO.inactivateStaleVolunteers(cutoffDate.toLocalDate(), switchToTimeZone);
		volunteerOrganizationDAO.inactivateForInactiveVolunteers();
		return updated;
	}

	@Override
	@PreAuthorize("hasAnyAuthority('" + Permission.VOLUNTEER_CREATE + ", " + Permission.LOGIN_KIOSK + "')")
	public void updatePreferredLanguage(long volunteerId, String language) {
		Language l = languageDAO.findByLanguageCode(language);
		if (l != null)
			volunteerDAO.updateFieldsWithoutVersionIncrement(volunteerId, false, null, false, null, null, null,
					l.getId());
	}

}
