package gov.va.vss.web.volunteer;

import static gov.va.shared.model.Permission.PermissionType.VOLUNTEER_CREATE;
import static gov.va.shared.util.SecurityUtil.hasAllPermissionsAtCurrentFacility;
import static gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType.ACTIVE;
import static gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType.TERMINATED;
import static gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType.TERMINATED_WITH_CAUSE;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.stream.Collectors;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.Permission;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.service.validation.ServiceValidationException;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.Donor;
import gov.va.vss.model.ParkingSticker;
import gov.va.vss.model.ParkingSticker.ParkingStickerView;
import gov.va.vss.model.ShirtSize;
import gov.va.vss.model.Uniform;
import gov.va.vss.model.Uniform.UniformView;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.Volunteer.VolunteerView;
import gov.va.vss.model.VolunteerAssignment;
import gov.va.vss.model.VolunteerAssignment.VolunteerAssignmentView;
import gov.va.vss.model.VolunteerOrganization;
import gov.va.vss.model.VolunteerOrganization.CompareByOrganization;
import gov.va.vss.model.VolunteerOrganization.VolunteerOrganizationView;
import gov.va.vss.model.benefitingService.BenefitingServiceRole;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.facility.Facility.FacilityView;
import gov.va.vss.model.lookup.RequirementDateType.RequirementDateTypeValue;
import gov.va.vss.model.lookup.RequirementStatus;
import gov.va.vss.model.lookup.VolunteerStatus;
import gov.va.vss.model.lookup.VolunteerStatus.VolunteerStatusType;
import gov.va.vss.model.organization.ScopeType;
import gov.va.vss.model.requirement.VolunteerRequirement;
import gov.va.vss.persistence.dao.TimeSummary;
import gov.va.vss.persistence.dao.VolunteerQuickSearchResult;
import gov.va.vss.persistence.dao.VolunteerQuickSearchResult.VolunteerQuickSearchResultView;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;
import gov.va.vss.persistence.queryCustomization.fieldTypes.VolunteerAssignmentAssociationFieldType;
import gov.va.vss.persistence.queryCustomization.fieldTypes.benefitingService.BenefitingServiceRoleFieldType;
import gov.va.vss.util.DateUtil;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.breadcrumbs.Breadcrumb;
import gov.va.vss.web.validation.ValidationException;
import gov.va.vss.web.volunteer.VolunteerSearchResults.Parameters;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class VolunteerController extends AbstractAppController {

	public static final String SESSION_ATTR_VOLUNTEER_SEARCH_MOST_RECENT = "volunteerSearchMostRecent";

	@Autowired
	private VolunteerValidator volunteerValidator;

	@RequestMapping("/getVolunteerSearchStations")
	@JsonView(FacilityView.Basic.class)
	public @ResponseBody SortedSet<Facility> getStationsForVolunteerSearch() {
		// Set<VAFacility> validFacilities =
		// SecurityUtil.getAllFacilitiesWhereUserHasPermissions(VOLUNTEER_READ);
		// return new TreeSet<>(validFacilities);
		return facilityDAO.findAllSorted(); // findVAFacilitiesWithLinkToFacility();
	}

	@RequestMapping("/volunteerOrganizations")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerOrganizationView.SearchForOrganizations.class)
	public @ResponseBody Map<String, Object> findOrgsForVolunteer(@RequestParam long volunteerId) {
		Map<String, Object> results = new HashMap<>();
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		SortedSet<VolunteerOrganization> r = new TreeSet<>(new CompareByOrganization());
		r.addAll(v.getVolunteerOrganizations());
		results.put("organizations", r);

		Map<Long, Double> hoursByOrganization = workEntryDAO.countByVolunteerAndBasicOrganizations(volunteerId,
				r.stream().map(p -> p.getOrganization().getId()).collect(Collectors.toList()));
		results.put("hoursByOrganization", hoursByOrganization);

		results.put("primaryOrganization", v.getPrimaryOrganization());
		return results;
	}

	@RequestMapping(value = "/volunteer/setPrimaryOrganization", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean primaryOrganizationSet(@RequestParam long volunteerId,
			@RequestParam long organizationId) {
		volunteerService.setPrimaryOrganization(volunteerId, organizationId);
		return true;
	}

	@RequestMapping(value = "/volunteer/deleteParkingSticker", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean parkingStickerDelete(@RequestParam long parkingStickerId) {
		parkingStickerService.delete(parkingStickerId);
		return true;
	}

	@RequestMapping(value = "/volunteer/deleteUniform", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean uniformDelete(@RequestParam long uniformId) {
		uniformService.delete(uniformId);
		return true;
	}

	@RequestMapping(value = "/volunteer/inactivateVolunteerAssignment", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Map<String, Object> volunteerAssignmentInactivate(@RequestParam long volunteerAssignmentId) {
		boolean statusChanged = volunteerService.inactivateAssignment(volunteerAssignmentId);
		Map<String, Object> result = new HashMap<>();
		result.put("volunteerStatusChanged", statusChanged);
		return result;
	}

	@RequestMapping(value = "/volunteer/deleteVolunteerAssignment", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Map<String, Object> volunteerAssignmentDelete(@RequestParam long volunteerAssignmentId,
			@ModelAttribute(DEFAULT_COMMAND_NAME) VolunteerCommand command) {
		boolean statusChanged = volunteerService.deleteAssignment(volunteerAssignmentId);
		/*
		 * Without this, Hibernate gets confused since the Volunteer in the
		 * command has a reference to a VolunteerAssignment that was already
		 * deleted - CPB
		 */
		command.setVolunteer(volunteerDAO.findRequiredByPrimaryKey(command.getVolunteer().getId()));

		Map<String, Object> result = new HashMap<>();
		result.put("volunteerStatusChanged", statusChanged);
		return result;
	}

	@RequestMapping(value = "/volunteer/volunteerRequirement/update", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean volunteerRequirementUpdate(@RequestParam long volunteerRequirementId,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam(required = false) LocalDate requirementDate,
			@RequestParam(required = false) RequirementStatus status, @RequestParam(required = false) String comments) {

		VolunteerRequirement vr = volunteerRequirementDAO.findRequiredByPrimaryKey(volunteerRequirementId);
		vr.setRequirementDate(requirementDate);
		vr.setStatus(status);
		vr.setComments(comments);

		vr = volunteerRequirementService.saveOrUpdate(vr);
		return true;
	}

	@RequestMapping(value = "/volunteer/inactivateOrganization", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Map<String, Object> organizationInactivate(@RequestParam long volunteerOrganizationId) {
		Map<String, Object> result = volunteerService.inactivateOrganization(volunteerOrganizationId);
		return result;
	}

	@RequestMapping(value = "/volunteer/deleteOrganization", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Map<String, Object> organizationDelete(@RequestParam long volunteerOrganizationId,
			@ModelAttribute(DEFAULT_COMMAND_NAME) VolunteerCommand command) {
		Map<String, Object> result = volunteerService.deleteOrganization(volunteerOrganizationId);
		/*
		 * Without this, Hibernate gets confused since the Volunteer in the
		 * command has a reference to a VolunteerOrganization that was already
		 * deleted - CPB
		 */
		command.setVolunteer(volunteerDAO.findRequiredByPrimaryKey(command.getVolunteer().getId()));
		return result;
	}

	@RequestMapping(value = "/volunteer/addOrganization", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Map<String, Object> organizationAdd(@RequestParam long volunteerId,
			@RequestParam long organizationId) throws ServiceValidationException {
		boolean statusChanged = volunteerService.addOrReactivateOrganization(volunteerId, organizationId);

		Map<String, Object> result = new HashMap<>();
		result.put("volunteerStatusChanged", statusChanged);
		return result;
	}

	@RequestMapping(value = "/volunteer/addOrReactivateVolunteerAssignment", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Map<String, Object> volunteerAssignmentAddOrReactivate(
			// either
			@RequestParam(required = false) Long volunteerAssignmentId,
			// or
			@RequestParam(required = false) Long volunteerId,
			@RequestParam(required = false) Long benefitingServiceRoleId) throws ServiceValidationException {
		long facilityId = getFacilityContextId();
		boolean statusChanged = volunteerService.addOrReactivateAssignment(volunteerAssignmentId, volunteerId,
				facilityId, benefitingServiceRoleId);
		Map<String, Object> result = new HashMap<>();
		result.put("volunteerStatusChanged", statusChanged);
		return result;
	}

	@RequestMapping(value = "/volunteer/parkingSticker/createOrUpdate", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean parkingStickerCreateOrUpdate(@RequestParam long facilityId,
			@RequestParam(required = false) Long parkingStickerId, @RequestParam long volunteerId,
			@RequestParam String number, @RequestParam(required = false) State state,
			@RequestParam(required = false) String licensePlate) {
		ParkingSticker ps;
		if (parkingStickerId != null) {
			ps = parkingStickerDAO.findRequiredByPrimaryKey(parkingStickerId);
		} else {
			ps = new ParkingSticker();
			Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
			ps.setVolunteer(v);
		}

		ps.setStickerNumber(number);
		ps.setLicensePlate(licensePlate);
		ps.setState(state);

		Facility i = facilityDAO.findRequiredByPrimaryKey(facilityId);
		ps.setFacility(i);

		ps = parkingStickerService.saveOrUpdate(ps);
		return true;
	}

	@RequestMapping(value = "/volunteer/setPrimaryFacility", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean primaryFacilitySet(@RequestParam long volunteerId, @RequestParam long facilityId) {
		volunteerService.setPrimaryFacility(volunteerId, facilityId);
		return true;
	}

	@RequestMapping("/findAvailableAssignments")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerAssignmentView.Search.class)
	public @ResponseBody List<AvailableAssignment> findAvailableAssignmentsForFacility(@RequestParam long facilityId) {
		List<BenefitingServiceRole> list = benefitingServiceRoleDAO.findByCriteria(null, Arrays.asList(facilityId),
				true, true, new QueryCustomization(BenefitingServiceRoleFieldType.BENEFITING_SERVICE));
		List<AvailableAssignment> results = new ArrayList<>();
		for (int i = 0; i < list.size(); i++) {
			BenefitingServiceRole r = list.get(i);
			AvailableAssignment a = new AvailableAssignment(i, r.getBenefitingService(), r);
			results.add(a);
		}
		return results;
	}

	@RequestMapping("/volunteerAssignments")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerAssignmentView.Search.class)
	public @ResponseBody Map<String, Object> findAssignmentsForVolunteer(@RequestParam long volunteerId) {
		Map<String, Object> results = new HashMap<>();
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		List<VolunteerAssignment> resultList = volunteerAssignmentDAO.findByCriteria(volunteerId, null, null, null,
				null, new QueryCustomization(VolunteerAssignmentAssociationFieldType.BENEFITING_SERVICE));
		results.put("assignments", resultList);

		Map<Long, Integer> hoursByAssignment = workEntryDAO
				.countByVolunteerAssignmentIds(resultList.stream().map(p -> p.getId()).collect(Collectors.toList()));
		results.put("hoursByAssignment", hoursByAssignment);

		results.put("primaryFacility", v.getPrimaryFacility());
		return results;
	}

	@RequestMapping("/volunteerRequirements")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	// @JsonView(VolunteerRequirementView.Search.class) __JC_TODO__
	public @ResponseBody Map<String, Object> findRequirementsForVolunteer(@RequestParam long volunteerId) {
		Map<String, Object> results = new HashMap<>();

		List<VolunteerRequirement> resultList = volunteerRequirementDAO.findByCriteria(volunteerId,
				getFacilityContextId());
		results.put("requirements", resultList);

		return results;
	}

	@RequestMapping("/volunteerParkingStickers")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(ParkingStickerView.Extended.class)
	public @ResponseBody SortedSet<ParkingSticker> findParkingStickersForVolunteer(@RequestParam long volunteerId) {
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
		SortedSet<ParkingSticker> r = new TreeSet<>();
		r.addAll(v.getParkingStickers());
		return r;
	}

	@RequestMapping("/volunteerUniforms")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(UniformView.Extended.class)
	public @ResponseBody SortedSet<Uniform> findUniformsForVolunteer(@RequestParam long volunteerId) {
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
		SortedSet<Uniform> r = new TreeSet<>();
		r.addAll(v.getUniforms());
		return r;
	}

	@RequestMapping("/volunteerSearch/find")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerView.Search.class)
	public @ResponseBody SortedSet<Volunteer> findVolunteers(@RequestParam(required = false) String firstName,
			@RequestParam(required = false) String lastName, @RequestParam(required = false) String code,
			@RequestParam(required = false) String email,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam(required = false) LocalDate dob,
			@RequestParam String scope, @RequestParam(required = false) Long facilityId,
			@RequestParam boolean includeInactive, HttpSession session) {
		if (StringUtil.allBlank(firstName, lastName, code, email) && dob == null)
			throw new IllegalArgumentException("Please specify at least one piece of search criteria");

		if ("National".equals(scope))
			facilityId = null;

		SortedSet<Volunteer> volunteers = new TreeSet<>(volunteerDAO.findByCriteria(firstName, null, lastName,
				code, dob, null, null, null, null, null, email, includeInactive ? null : VolunteerStatusType.ACTIVE,
				facilityId != null ? Arrays.asList(facilityId) : null));

		session.setAttribute(SESSION_ATTR_VOLUNTEER_SEARCH_MOST_RECENT,
				new VolunteerSearchResults(volunteers, getFacilityContextId(),
						new Parameters(firstName, lastName, code, email, dob, scope, facilityId, includeInactive)));

		return volunteers;
	}

	@RequestMapping("/volunteerSearch/mostRecent")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerView.Search.class)
	public @ResponseBody VolunteerSearchResults volunteerSearchMostRecent(HttpSession session) {
		VolunteerSearchResults r = (VolunteerSearchResults) session
				.getAttribute(SESSION_ATTR_VOLUNTEER_SEARCH_MOST_RECENT);
		if (r == null)
			return null;

		if (r.facilityId != getFacilityContextId()) {
			session.removeAttribute(SESSION_ATTR_VOLUNTEER_SEARCH_MOST_RECENT);
			return null;
		}

		return r;
	}

	@RequestMapping("/volunteerDuplicateCheck")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerView.Search.class)
	public @ResponseBody SortedSet<Volunteer> findDuplicateVolunteers(@RequestParam String firstName,
			@RequestParam String lastName,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate dob,
			@RequestParam(required = false) Long excludeVolunteerId) {
		if (StringUtils.isBlank(lastName) || StringUtils.isBlank(firstName))
			throw new IllegalArgumentException("Both last name and first name are required.");

		SortedSet<Volunteer> volunteers = new TreeSet<>();
		List<Volunteer> firstNameMatches = volunteerDAO.findByCriteria(firstName, null, null, null, dob, null, null,
				null, null, null, null, null, null);
		volunteers.addAll(firstNameMatches);

		List<Volunteer> lastNameMatches = volunteerDAO.findByCriteria(null, null, lastName, null, dob, null, null, null,
				null, null, null, null, null);
		volunteers.addAll(lastNameMatches);

		if (excludeVolunteerId != null) {
			volunteers = volunteers.stream().filter(p -> !p.getId().equals(excludeVolunteerId))
					.collect(Collectors.toCollection(TreeSet::new));
		}

		return volunteers;
	}

	@RequestMapping("/volunteerCreate.htm")
	// Don't want a Breadcrumb here since we want to force them to search first
	// every time - CPB
	// @Breadcrumb("Create Volunteer")
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public String volunteerCreate(ModelMap model, HttpServletRequest request) {
		Volunteer volunteer = new Volunteer();
		/*
		 * Will be overwritten during the saveOrUpdate; this prevents the UI
		 * from displaying this initially - CPB
		 */
		volunteer.setStatus(volunteerStatusDAO.findByLookup(ACTIVE));
		volunteer.setMealsEligible(1);
		volunteer.setStatusDate(getTodayAtFacility());

		VolunteerCommand command = new VolunteerCommand(volunteer, null, null);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(command, model);
		return "createVolunteer";
	}

	@RequestMapping("/volunteerEdit.htm")
	@Breadcrumb("Edit Volunteer")
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public String volunteerEdit(@RequestParam long id, @RequestParam(required = false) String fromPage, ModelMap model,
			HttpServletRequest request) {
		Volunteer volunteer = volunteerDAO.findRequiredByPrimaryKey(id);

		TimeSummary summary = volunteerDAO.getTimeSummary(id, getFacilityTimeZone());
		VolunteerCommand command = new VolunteerCommand(volunteer, summary, fromPage);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		model.addAttribute("DateValueNotApplicable", RequirementDateTypeValue.NOT_APPLICABLE);
		// WebUtil.addClassConstantsToModel(RequirementDateType.class, model);
		// WebUtil.addEnumToModel(RequirementDateTypeValue.class, model);

		createReferenceData(command, model);

		return "editVolunteer";
	}

	private void createReferenceData(VolunteerCommand command, ModelMap model) {
		Volunteer volunteer = command.getVolunteer();
		model.put("allGenders", genderDAO.findAllSorted());
		model.put("allLanguages", languageDAO.findAllSorted());
		model.put("allTransportationMethods", transportationMethodDAO.findAllSorted());
		model.put("allAwards", awardDAO.findAllSorted());
		model.put("allShirtSizes", uniformDAO.findAllShirtSizes());
		model.put("allRequirementStatuses", requirementStatusDAO.findAllSorted());

		Set<VAFacility> editSites = getCurrentUser()
				.getFacilitiesWhereUserHasAllPermissions(PermissionType.VOLUNTEER_CREATE);
		Map<Long, Facility> i = facilityDAO.findByVAFacilities(editSites);
		List<Long> facilityIds = i.values().stream().filter(p -> !p.isCentralOffice()).map(p -> p.getId())
				.collect(Collectors.toList());
		model.addAttribute("volunteerEditSiteIds", StringUtils.join(facilityIds, ','));

		model.addAttribute("backgroundStatus", "No Issue Reported");
		if (volunteer.isPersistent()) {
			model.addAttribute("allVolunteerFacilities", volunteerDAO.findFacilitiesForVolunteer(volunteer.getId()));
			model.addAttribute("yearsVolunteering", workEntryDAO.getNumYearsWorked(volunteer.getId()));
		}

		model.addAttribute("volunteerHasActiveAssignmentAtCurrentFacility", volunteer.getVolunteerAssignments().stream()
				.anyMatch(p -> p.isActive() && p.getFacility().getRootFacilityId() == getFacilityContextId()));
		model.addAttribute("volunteerHasActiveOrganizationAtCurrentFacility",
				volunteer.getVolunteerOrganizations().stream()
						.anyMatch(p -> p.isActive() && (p.getOrganization().getScope() == ScopeType.NATIONAL
								|| p.getOrganization().getFacility().getRootFacilityId() == getFacilityContextId())));
		if (!hasAllPermissionsAtCurrentFacility(VOLUNTEER_CREATE) || command.getTerminationDate() != null)
			setFormAsReadOnly(model, true);
		if (hasAllPermissionsAtCurrentFacility(VOLUNTEER_CREATE) && command.getTerminationDate() != null)
			model.addAttribute("enableTerminationControls", true);
	}

	@RequestMapping("/volunteerSubmit.htm")
	@PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_CREATE + "')")
	public String volunteerSubmit(@ModelAttribute(DEFAULT_COMMAND_NAME) VolunteerCommand command, BindingResult result,
			SessionStatus status, ModelMap model, HttpServletRequest request) throws ValidationException {
		Volunteer volunteer = command.getVolunteer();

		boolean isEdit = volunteer.isPersistent();

		volunteerValidator.validate(command, result, false, "volunteer");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {

			if (command.getTerminationDate() != null) {
				volunteer.setStatus(volunteerStatusDAO
						.findByLookup(command.isVolunteerTerminatedWithCause() ? TERMINATED_WITH_CAUSE : TERMINATED));
				volunteer.setStatusDate(command.getTerminationDate());
			} else {
				VolunteerStatus newStatus = volunteerStatusDAO.findByPrimaryKey(volunteer.getStatus().getId());
				if (command.getTerminationDate() == null && newStatus.getLookupType().isTerminated()) {
					/*
					 * saveOrUdpate call below will reset this to INACTIVE if
					 * necessary - CPB
					 */
					volunteer.setStatus(volunteerStatusDAO.findByLookup(ACTIVE));
					volunteer.setStatusDate(getTodayAtFacility());
				}
			}
			volunteer.setTerminationRemarks(command.getTerminationRemarks());

			try {
				volunteer = volunteerService.saveOrUpdate(volunteer, getCurrentUser());
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "volunteer.update.success" : "volunteer.create.success"));
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(command, model);
			return isEdit ? "editVolunteer" : "createVolunteer";
		} else {
			status.setComplete();

			String toPage = "/volunteerEdit.htm?id=" + volunteer.getId();
			if ("donor".equals(command.getFromPage())) {
				Donor donor = volunteer.getDonor();
				if (donor != null)
					toPage = "/donorEdit.htm?id=" + donor.getId();
			}
			return "redirect:" + toPage;
		}
	}

	@RequestMapping(value = "/volunteer/uniform/createOrUpdate", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean uniformCreateOrUpdate(@RequestParam long facilityId,
			@RequestParam(required = false) Long uniformId, @RequestParam long volunteerId,
			@RequestParam(required = false) ShirtSize size, @RequestParam(required = false) Integer count) {
		Uniform u;
		if (uniformId != null) {
			u = uniformDAO.findRequiredByPrimaryKey(uniformId);
		} else {
			u = new Uniform();
			Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
			u.setVolunteer(v);
		}

		u.setShirtSize(size);
		u.setNumberOfShirts(count);

		Facility i = facilityDAO.findRequiredByPrimaryKey(facilityId);
		u.setFacility(i);

		u = uniformService.saveOrUpdate(u);
		return true;
	}

	@RequestMapping("/volunteer/quickSearch/currentFacility")
	@JsonView(VolunteerQuickSearchResultView.TimeEntrySearch.class)
	@PreAuthorize("hasAnyAuthority('" + Permission.TIME_CREATE + ", " + Permission.TIME_READ + "')")
	public @ResponseBody Map<String, Object> quickSearchVolunteersForCurrentFacility(
			@RequestParam(required = false) String name) {
		long facilityId = getFacilityContextId();

		Map<String, Object> resultMap = new HashMap<>();

		SortedSet<VolunteerQuickSearchResult> quickSearch = volunteerDAO.quickSearch(name, null, facilityId, true, true,
				true);

		/*
		 * Only show volunteers that have at least one active assignment and
		 * organization - CPB
		 */
		SortedSet<VolunteerQuickSearchResult> results = quickSearch.stream()
				.filter(p -> !p.getAssignments().isEmpty() && !p.getOrganizations().isEmpty())
				.collect(Collectors.toCollection(TreeSet::new));

		resultMap.put("volunteers", results);

		return resultMap;
	}

	@RequestMapping("/volunteer/quickSearch/individualPlusAssignmentsAndOrgs")
	@JsonView(VolunteerQuickSearchResultView.TimeEntrySearch.class)
	@PreAuthorize("hasAnyAuthority('" + Permission.TIME_CREATE + ", " + Permission.TIME_READ + "')")
	public @ResponseBody VolunteerQuickSearchResult volunteerWithAssignmentsAndOrgsAtCurrentFacility(
			@RequestParam long volunteerId, @RequestParam boolean onlyActiveAssignmentsAndOrgs) {
		long facilityId = getFacilityContextId();
		SortedSet<VolunteerQuickSearchResult> results = volunteerDAO.quickSearch(null, volunteerId, facilityId, true,
				true, onlyActiveAssignmentsAndOrgs);
		return results.isEmpty() ? null : results.iterator().next();
	}

}
